# The default reviewer
# The parent module (reviewers/__init__.py) contatins some
# generic code used by all of the reviewers.
from vinca.lib import ansi
import readchar 

from vinca.lib.video import DisplayImage
from vinca.lib.audio import Recording
from vinca.lib.terminal import AlternateScreen

def make_string(card):
	f = ' / '.join((card.path / 'front').read_text().splitlines())
	b = ' / '.join((card.path / 'back').read_text().splitlines())
	return f + ' | ' + b

def review(card):
	with AlternateScreen():
		# front text
		front = (card.path / 'front').read_text()
		print(front)
		tags_str = '    '.join(['['+t.replace('_',' ')+']' for t in card.tags])  # TODO show tags only if requested
		ansi.light()
		print(tags_str)
		ansi.reset()

		# front media
		with DisplayImage(card.path/'image_front'), Recording(card.path/'audio_front'):
			# card flip
			readchar.readchar() 

		back = '\n' + (card.path / 'back').read_text()
		print(back)

		with DisplayImage(card.path/'image_back'), Recording(card.path/'audio_back'):
			key = readchar.readchar()
	
	return key



import readline as rl

from vinca.lib import ansi
from vinca.lib.terminal import COLUMNS

def edit(card, scrollback = True):
	front_path, back_path = (card.path / 'front'), (card.path / 'back')
	front = front_path.read_text()
	front.replace('\n',';')
	rl.set_startup_hook(lambda: rl.insert_text(front))
	new_front = input('Q: ').replace(';','\n')
	front_path.write_text(new_front)
	
	back = back_path.read_text().replace('\n',';')
	rl.set_startup_hook(lambda: rl.insert_text(back))
	new_back = input('A: ').replace(';','\n')
	back_path.write_text(new_back)

	rl.set_startup_hook()
	if scrollback:
		ansi.up_line(2)




# The metareviewer which implements specific reviewers
import importlib
import time, datetime

TODAY = datetime.date.today()
	
GRADE_DICT = {'x': 'delete', 'd': 'delete',
	      'q': 'exit', '\x1b': 'exit',
	      '1': 'again',
	      '2': 'hard',
	      '3': 'good', ' ': 'good', '\r': 'good', '\n': 'good',
	      '4': 'easy'}

def review(card):

	start = time.time()

	# dynamically import the required reviewer module
	# a specifc reviewer is responsible for returning a key to the generic reviewer
	m = importlib.import_module('.'+card.reviewer, package = 'vinca.reviewers')
	key = m.review(card)  # the reviewer gives back the key

	stop = time.time()
	elapsed = int(stop - start)

	grade = GRADE_DICT[key] if key in GRADE_DICT else 'exit'

	card.add_history(TODAY, elapsed, grade)

def make_string(card):
	m = importlib.import_module('.'+card.reviewer, package = 'vinca.reviewers')
	assert hasattr(m, 'make_string'), f'{card.reviewer} must implement \
		the make_string method to represent this card on the command line'
	return m.make_string(card)
	



import importlib

def schedule(card):

	if card.last_grade == 'delete':
		card.deleted = True

	# import the specific scheduler module
	m = importlib.import_module('.' + card.scheduler, package = 'vinca.schedulers')
	# invoke the specific scheduler
	m.schedule(card)



import tkinter as Tk
from subprocess import run
from pathlib import Path

TERMINAL_BACKGROUND = '#000000'

class ActiveWindow:
	def __init__(self):
		# use xdotool to query basic info about the terminal's geometry
		out = run(['xdotool','getactivewindow','getwindowgeometry'], capture_output=True)
		# parse output and calculate references
		out = str(out.stdout, encoding='utf-8')
		window_id, abs_pos, geometry = out.splitlines()
		self.left, self.top = abs_pos.split()[1].split(',')
		self.left, self.top = int(self.left), int(self.top)
		self.width, self.height = geometry.split()[1].split('x')
		self.width, self.height = int(self.width), int(self.height)
		self.right = self.left + self.width
		self.bottom = self.top + self.height
		self.center_x = self.left + self.width // 2
		self.center_y = self.top + self.height // 2

class DisplayImage:
	'''A simple class to draw an image to the screen.
	There are two methods: show and close.
	It can also be invoked as a context manager.'''

	def __init__(self, image_path):
		self.image_path = image_path

	def show(self):
	# get geometry information about the active terminal
		aw = ActiveWindow()

		self.root = Tk.Tk()
		self.image = Tk.PhotoImage(file=self.image_path)
		# warning: images cannot be stored as local variables
		# of a function or else there is risk of them being
		# destroyed by premature garbage collection.

		# we want to center the image at the bottom of the terminal
		# with a 40 pixel margin on all sides
		# if the image is too big we will only see part of it
		# but the window will fit inside the active terminal
		img_height = self.image.height()
		img_width = self.image.width()
		margin = 40
		left = max(aw.left + margin, aw.center_x - img_width // 2)
		right = min(aw.right - margin, aw.center_x + img_width // 2)
		width = right - left
		bottom = aw.bottom - margin
		top = max(aw.top + margin, bottom - img_height)
		height = bottom - top
		self.root.geometry(f'{width}x{height}+{left}+{top}')

		# we do not want to let the window manager make decisions
		self.root.overrideredirect(True)

		# draw the image on a canvas
		# place our canvas to occupy the whole window
		self.canvas = Tk.Canvas(self.root, width=width, height=height)
		self.canvas.config(bg=TERMINAL_BACKGROUND)
		self.canvas.create_image(0,0, anchor=Tk.NW, image=self.image)
		self.canvas.place(x=-1,y=-1, height=height+2, width=width+2)

		# manually draw our window to the screen
		# it is common to see root.mainloop(), but
		# that is unneeded as there is not UX here.
		self.root.update()

	def close(self):
		assert self.root
		self.root.destroy()

	def __enter__(self):
		if self.image_path.exists():
			self.show()

	def __exit__(self, *exception_args):
		if self.image_path.exists():
			self.close()



import sys
import inspect
import readchar
import readline as rl
import argparse
import datetime
from pathlib import Path
from shutil import copytree, rmtree

from vinca.lib.classes import Card
from vinca.lib import ansi
from vinca.lib import filter
from vinca.generators import generate, GENERATORS_DICT

# TODO: intelligent scrollback for reviewers and editors
# TODO: make functions specify the parameters they need and 
# feed them this explicitly á la FILTER

TODAY = datetime.date.today()

COLORIZE = sys.stdout.isatty()
QUIT_KEYS = ['q', readchar.key.ESC, '\n', '\r']

vinca_path = Path(__file__).parent 
cards_path = vinca_path / 'cards'
ALL_CARDS = [Card(int(id.name)) for id in cards_path.iterdir()]
ALL_TAGS = [tag for card in ALL_CARDS for tag in card.tags] # TODO: cache

# ARGUMENT PARSING
parser = argparse.ArgumentParser(usage=argparse.SUPPRESS, epilog='Use -h to see help for any command. Visit the manual for usage examples.')
subparsers = parser.add_subparsers(metavar='')
def card_type(arg):
	id = int(arg.split()[0])  # grab the first field of the argument
	return Card(id)
def date_type(s):
	if s[s[0] in ['+','-']:].isdigit():
		return TODAY + int(arg) * DAY
	try:
		return datetime.datetime.strptime(s, '%Y-%m-%d').date()
	except ValueError:
		raise argparse.ArgumentTypeError(f'''\n\n
			Invalid Date: {s}. Valid dates are:
			1) -7		(one week ago)
			2) 2021-06-03	(June 3rd)''')
def arg(*names_or_flags, **kwargs):
	return names_or_flags, kwargs
CARDS = arg('cards',nargs='*',type=card_type, default=ALL_CARDS)
CARD = arg('card',type=card_type)
OPTIONAL_CARD = arg('card', nargs='?', type=card_type)
OPTIONAL_CARDS = arg('--cards',nargs='*',type=card_type,default=ALL_CARDS)
OPTIONAL_CARDS_DEFAULT_NONE = arg('--cards',nargs='*',type=card_type,default=[])
def subcommand(*subparser_args, parent=subparsers, alias='', name=None):
	def decorator(func, alias=alias, name=name):
		name = name if name else func.__name__
		description = func.__doc__
		aliases = [alias] if alias else []
		parser = parent.add_parser(name, help='', description=description, aliases=aliases,
		                           usage=argparse.SUPPRESS)
		for args, kwargs in subparser_args:
			parser.add_argument(*args, **kwargs)
		parser.set_defaults(func=func)
		return func
	return decorator

@subcommand(arg('-t','--default_tags', nargs='+', default=[]), alias='a')
def add(args): 
	ansi.hide_cursor()
	print(*[f'{key}\t{generator}' for key, generator in GENERATORS_DICT.items()],sep='\n')
	k = readchar.readchar()
	if k not in GENERATORS_DICT: return
	args.generator = GENERATORS_DICT[k]
	args.cards = generate(args)
	ansi.up_line(len(GENERATORS_DICT) + 2)
	ansi.clear_to_end()
	browse(args)

@subcommand(OPTIONAL_CARD, OPTIONAL_CARDS, alias='s')
def statistics(args):
	if args.card:
		card = args.card
		print(f'\nCard #{card.id}')
		print(str(card))
		print(f'Tags: {" ".join(card.tags)}')
		print(f'Due: {card.due_date}')
		print('Date\t\tTime\tGrade')
		print(*[f'{date}\t{time}\t{grade}' for date, time, grade in card.history],sep='\n',end='')
		lines = 5+len(card.history)
	elif args.cards:
		due_cards = filter.filter(args.cards, due_only=True)
		new_cards = filter.filter(args.cards, new_only=True)
		print('Total', len(args.cards), sep='\t')
		print('Due', len(due_cards), sep='\t')
		print('New', len(new_cards), sep='\t')
		lines = 3
	if args.scrollback:
		ansi.up_line(lines)

@subcommand(CARD,alias='e')
def edit(args):
	''' edit a single card '''
	args.card.edit()

@subcommand(CARD,alias='E')
def edit_metadata(args):
	''' edit the metadata of a card '''
	args.card.edit_metadata()

@subcommand(OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE, alias='x')
def delete(args):
	if args.card:
		args.card.deleted = not args.card.deleted
	elif args.cards:
		for card in args.cards:
			card.deleted = not card.deleted

@subcommand(CARDS, alias='r')
def review(args):
	if args.card:
		args.card.review()
		args.card.schedule()
	elif args.cards:
		args.cards = filter.filter(args.cards, due_only = True)
		if not args.cards:
			print('No cards due.')
			return
		browse(args, reviewing = True)

@subcommand(arg('tag'), OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE)
def add_tag(args):
	for card in [args.card] if args.card else args.cards:
		card.tags += [args.tag]

@subcommand(arg('tag'), OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE)
def remove_tag(args):
	for card in [args.card] if args.card else args.cards:
		if args.tag in card.tags:
			card.tags.remove(args.tag)
		# TODO do this with set removal
		card.save_metadata()

@subcommand(OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE, alias='t')
def edit_tags(args):
	def complete(text, state):
		for tag in ALL_TAGS:
			if tag.startswith(text):
				if not state:
					return tag
				else:
					state -= 1
	rl.parse_and_bind('tab: complete')
	rl.set_completer(complete)

	if args.card:
		pass #TODO lineedit tags with tab complete
		tags = ' '.join(args.card.tags)
		rl.set_startup_hook(lambda: rl.insert_text(tags))
		args.card.tags = input('tags: ').split()
		if args.scrollback:
			ansi.up_line()
		rl.set_startup_hook()
	elif args.cards:
		tags_add = input('tags to add: ').split()
		tags_remove = input('tags to remove: ').split()
		for tag in tags_add:
			args.tag = tag
			add_tag(args)
		for tag in tags_remove:
			args.tag = tag
			remove_tag(args)

	rl.set_completer()

CMD_DICT = {'r': review, 'R': review,
	    's': statistics, 'S': statistics,
	    'x': delete, 'X': delete,
	    'e': edit, 'E': edit_metadata,
	    't': edit_tags, 'T': edit_tags}
@subcommand(CARDS, arg('-t','--default_tags', nargs='+', default=[]), alias='b')
def browse(args, reviewing = False):
	args.scrollback = True
	cards = args.cards
	# TODO max frame of ten cards
	n = len(cards); sel = 0

	def quit():
		ansi.show_cursor()
		ansi.line_wrap_on()
		ansi.clear_to_end()
		exit()
	def draw_browser():
		ansi.hide_cursor()
		ansi.line_wrap_off()

		for i, card in enumerate(cards):
			if card.due_as_of(args.review_date):
				ansi.bold()
				ansi.blue()
			if card.deleted:
				ansi.crossout()
				ansi.red()
			if i==sel:
				ansi.highlight()
			print(card)
			ansi.reset()

	def redraw_browser():
		ansi.up_line(n)
		ansi.clear_to_end()
		draw_browser()
		
	draw_browser()
	while True:

		k = 'R' if reviewing else readchar.readchar()

		sel += k=='j'
		sel -= k=='k'
		sel %= n

		if k in QUIT_KEYS:
				quit()
		if k in CMD_DICT:
			ansi.show_cursor()
			args.card = cards[sel] if k in ('x','s','r','R','e','E', 't') else None
			CMD_DICT[k](args)
			ansi.hide_cursor()
			reviewing = (k == 'R' and cards[sel].last_grade != 'exit')
			if reviewing and sel == n - 1:
				quit()
			elif reviewing and sel < n - 1:
				sel += 1 
		if k in GENERATORS_DICT:
			ansi.show_cursor()
			args.generator = GENERATORS_DICT[k]
			new_cards = generate(args)
			cards[sel:sel] = new_cards
			n = len(cards)
			ansi.down_line(len(new_cards))
			ansi.hide_cursor()

		if k not in ('s','S'):
			redraw_browser()
			

@subcommand(arg('pattern',nargs='?',default=''),
	arg('-v','--invert',action='store_true'),
	arg('-i','--id_only',action='store_true'),
	arg('--cards',nargs='*',type=card_type, default=ALL_CARDS),
	arg('--tags_include',nargs='*', metavar='TAGS'),
	arg('--tags_exclude',nargs='*', metavar='TAGS'),
	arg('--create_date_min',type=date_type, metavar='DATE'),
	arg('--create_date_max',type=date_type, metavar='DATE'),
	arg('--seen_date_min',type=date_type, metavar='DATE'),
	arg('--seen_date_max',type=date_type, metavar='DATE'),
	arg('--due_date_min',type=date_type, metavar='DATE'),
	arg('--due_date_max',type=date_type, metavar='DATE'),
	arg('--due_only',action='store_true'),
	arg('--not_due_only',action='store_true'),
	arg('--editor', type=str),
	arg('--reviewer', type=str),
	arg('--scheduler', type=str),
	arg('--deleted_only',action='store_true'),
	arg('--show_deleted',action='store_true'),
	arg('--new_only',action='store_true'),
	arg('--not_new_only',action='store_true'),
	alias='f', name='filter')
def display_filter(args):
	''' Filter a set of cards using a variety of parameters.
	Piping the output of this \033[32m command to another vinca command will
	provide a list of cards.'''

	# get filter parameters as a list of strings
	filter_kwargs = inspect.getfullargspec(filter.filter).args[1:]
	# check that args specifies these parameters
	assert all([hasattr(args, param) for param in filter_kwargs])
	matches = filter.filter(args.cards,
		# feed the keyword args editor=args.editor, due=args.due, 
		**{param : getattr(args, param) for param in filter_kwargs})
	for card in matches:
		if args.id_only:
			print(card.id)
			continue
		if COLORIZE:
			if card.due_as_of(TODAY):
				ansi.bold()
				ansi.blue()
			if card.deleted:
				ansi.crossout()
				ansi.red()
		print(card.id, card, sep='\t')
		if COLORIZE:
			ansi.reset()
	
@subcommand()
def purge(args):
	''' Permanently delete all cards marked for deletion. '''
	for card in filter.filter(ALL_CARDS, deleted_only=True):
		rmtree(card.path)

@subcommand(arg('backup_path',type=Path), CARDS, name='export')
def exp(args):
	for card in args.cards:
		copytree(card.path, args.backup_path / str(card.id))

@subcommand(arg('import_path',type=Path), arg('-o','--overwrite', action='store_true'), name='import')
def imp(args):
	if args.overwrite:
		rmtree(cards_path)
		copytree(args.import_path, cards_path)
		return
	old_ids = [card.id for card in ALL_CARDS]
	for new_id,card_path in enumerate(args.import_path.iterdir(), max(old_ids, default=1) + 1):
		copytree(card_path, cards_path / str(new_id))


for alias, generator in GENERATORS_DICT.items():
	p = subparsers.add_parser(generator, aliases=[alias], help='')
	p.add_argument('-t','--default_tags',nargs='*',default=[])
	p.set_defaults(func = generate, generator = generator)

# parse the command line arguments
parser.set_defaults(cards = [], func = lambda args: parser.print_help(), review_date = TODAY, scrollback = False, card = None)
parser._action_groups[0].title = 'commands'
parser._action_groups.pop()
args = parser.parse_args()
# accept a file of newline separated card ids
if not sys.stdin.isatty():
	ids = [int(line.strip().split()[0]) for line in sys.stdin]
	args.cards = [Card(id) for id in ids]
	sys.stdin = open('/dev/tty')  
args.func(args)



from vinca.lib import classes

def generate(args):
	new_card = classes.Card(create=True)
	new_card.editor, new_card.reviewer, new_card.scheduler = 'verses', 'verses', 'base'
	new_card.edit()
	return [new_card]



from vinca.lib import classes
from vinca.lib import ansi

def generate(args):
	new_card = classes.Card(create=True)
	new_card.editor, new_card.reviewer, new_card.scheduler = 'two_lines', 'two_lines', 'base'
	front = input('Q:   ')
	back = input('A:   ')
	(new_card.path/'front').write_text(front)
	(new_card.path/'back').write_text(back)
	if args.scrollback:
		ansi.up_line(2)
	return [new_card]



# linear editor for lyrics, poetry, oratory, etc.
import subprocess
from pathlib import Path
import shutil

vinca_path = Path(__file__).parent.parent.parent # I counted right
tags_path = vinca_path / 'data' / 'tags.txt'

path = Path(__file__).parent
vimrc_path = path / 'vimrc'

def edit(card):
	# we are going to run vim...
	vim_cmd = ['vim']
	vim_cmd += [card.path/'lines']

	# we are going to open each file window
	# one by one so that we can specify things well
	# this tells vim that we do not want to automatically
	# resize/equalize windows each time we open one
	# l will store a list of commands to open the window splits
	l = ['set noequalalways']
	# when we open a window split below
	# we automatically move focus to that window
	l += ['set splitbelow']
	l += [f'1 split {card.path / "tags"}']
	# bring focus back up to the top
	l += ['1 wincmd W']
	vim_cmd += ['-c',' | '.join(l)]
	# using a vimrc file to make a few custom bindings...
	vim_cmd += ['-Nu', vimrc_path]
	# including tag autocompletion
	vim_cmd += ['-c', f'set dictionary={tags_path}']
	# launch
	subprocess.run(vim_cmd)



import re
import functools

# ansi codes let us control the terminal just
# by printing special escape sequences
esc = '\033['
codes = {'reset': esc+'0m',
        'bold': esc+'1m', 'b': esc+'1m',
        'bold_off': esc+'22m', 'b': esc+'1m',
        'light': esc+'2m', 'dim': esc+'2m',
        'italic': esc+'3m', 'it': esc+'3m',
        'underlined': esc+'4m', 'u': esc+'4m',
        'blink': esc+'5m', 'flash': esc+'5m', 'f': esc+'5m',
        'highlight': esc+'7m', 'hi': esc+'7m', 'reverse': esc+'7m',  'mark': esc+'7m',
        'hidden': esc+'8m', 'invisible': esc+'8m',
	'crossout': esc+'9m',
        'red': esc+'31m', 'r': esc+'31m',
        'green': esc+'32m', 'g': esc+'32m',
        'yellow': esc+'33m',
	'blue': esc+'34m',
	'reset_color': esc+'39m',
	# the next escape codes are more powerful
	'clear': esc+'2J',
	'clear_to_end': esc+'J',
	'move_to_top': esc+'H',  # home
	'hide_cursor': esc+'?25l',
	'show_cursor': esc+'?25h',
	'save_cursor': esc+'s',
	'restore_cursor': esc+'u',
	'save_screen': esc+'?47h',
	'restore_screen': esc+'?47l',
	'up_line': esc+'F',
	'down_line': '\033E',
	'clear_line': esc+'K',
	'line_wrap_off': esc+'?7l',
	'line_wrap_on': esc+'?7h',
	}

def _do(n=1, cmd=None):
	print(codes[cmd]*n, end='')
for key in codes:
	globals()[key] = functools.partial(_do, cmd=key)



import subprocess
import json
import datetime
TODAY = datetime.date.today()
from pathlib import Path

from vinca import reviewers, editors, schedulers 

vinca_path = Path(__file__).parent.parent  # /path/to/vinca
cards_path = vinca_path / 'cards'

class Card:
	# Card class can load without 
	default_metadata = {'editor': 'base', 'reviewer':'base', 'scheduler':'base',
			    'tags': [], 'history': [], 'deleted': False,
			    'due_date': TODAY, 'siblings': [], 'string': ''}

	def __init__(self, id=None, create=False):
		assert (id is not None) ^ create
		if not create:  # load an existing card
			assert type(id) is int
			self.id = id
			self.metadata_is_loaded = False
		elif create: # create a new card
			old_cids = [int(x.name) for x in cards_path.iterdir()]
			self.id = max(old_cids) + 1 if old_cids else 100 
			self.path.mkdir()
			self.metadata = Card.default_metadata
			self.metadata_is_loaded = True
			self.save_metadata()

	@property
	def path(self):
		return cards_path/str(self.id)
		# assert p.exists(), f'Bad Card ID {self.id}; DOES NOT EXIST.'

	@property
	def metadata_path(self):
		return self.path / 'metadata.json'

	def load_metadata(self):
		self.metadata = json.load(self.metadata_path.open())
		# dates must be serialized into strings for json
		# I unpack them here
		str_to_date = lambda date: datetime.datetime.strptime(date, '%Y-%m-%d').date()
		self.metadata['history'] = [[str_to_date(date), grade,
					    time] for date, grade, time in self.metadata['history']]
		self.metadata['due_date'] = str_to_date(self.metadata['due_date'])
		self.metadata_is_loaded = True

	def save_metadata(self):
		json.dump(self.metadata, self.metadata_path.open('w'), default=str, indent=2)

	for m in default_metadata.keys():
		# create getter and setter methods for everything in the metadata dictionary
		exec(f'''
@property
def {m}(self):
	if not self.metadata_is_loaded:
		self.load_metadata()
	return self.metadata["{m}"]''')
		exec(f'''
@{m}.setter
def {m}(self, new_val):
	if not self.metadata_is_loaded:
		self.load_metadata()
	self.metadata["{m}"] = new_val
	self.save_metadata()''')	

	def __str__(self):
		return self.string

	def add_history(self, date, time, grade):
		self.history = self.history + [[date, time, grade]]
	def undo_history(self):
		self.history = self.history[:-1]
			
	def review(self): reviewers.review(self)
	def make_string(self): self.string = reviewers.make_string(self)
	def edit(self):
		editors.edit(self) 
		self.make_string()
	def edit_metadata(self):
		subprocess.run(['vim',self.path/'metadata.json'])
		self.load_metadata()
		self.make_string()
	def schedule(self): schedulers.schedule(self) 

	@property
	def create_date(self): return self.history[0][0]
	@property
	def seen_date(self): return self.history[-1][0]
	@property
	def last_grade(self): return self.history[-1][2]
	@property
	def last_interval(self): return self.history[-1][0] - self.history[-2][0]  

	@property
	def new(self): return self.last_grade == 'create'

	def due_as_of(self, date):
		return self.due_date <= date




import setuptools



setuptools.setup(
    name="vinca",
    version="1.1.98",
    author="Oscar Laird", 
    include_package_data = True,
    packages=setuptools.find_packages(),
    install_requires=[ 
        'readchar',
    ],
    scripts = ['scripts/vinca','scripts/vinca_debug']
)



# try vlc or playsound modules (or mixer from pygame)

class Recording:
	''' There are two methods: play and stop.
	It can also be invoked as a context manager.'''

	def __init__(self, audio_path):
		self.audio_path = audio_path

	def play(self):
		pass

	def stop(self):
		pass


	def __enter__(self):
		if self.audio_path.exists():
			self.play()

	def __exit__(self, *exception_args):
		if self.audio_path.exists():
			self.stop()
